/* ifile - intelligent mail filter for EXMH/MH
   Copyright (C) 1997  Jason Rennie <jr6b+@andrew.cmu.edu>
   
   ifilter.c code Copyright (C) 1997 by Jason Rennie <jr6b+@andrew.cmu.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program (see file 'COPYING'); if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
   */

/*
  ifilter.c
  
  C implementation of ifilter.perl, written by Jason Rennie.
  
  C code primarily written by Diego Zamboni <zamboni@cs.purdue.edu>, Mar 7, 1997
     additions and changes by Jason Rennie <jr6b@cs.cmu.edu>

  No attempt at modification of the algorithm of program was
  made. This reimplementation was made solely with the attempt of
  making the program run faster. I receive a lot of mail, and
  filtering it just takes too much time.
  
  Some optimizations were made, particularily in the analysis of the
  message. Some of the most important ones are:
  
  - The message is "pre-partitioned" in words and the words are stored
    in an array which is then used repeatedly for computing the target
    value for each folder. This is _much_ faster than repartitioning
    the entire message for each category.

  I am still thinking about some optimizations that could be made to
  the database-reading part. But for now, the filtering time is down
  to ~3 sec, from ~17 sec previously. Also, the memory footprint of
  this program is _much_ smaller than the perl one, which makes its
  execution faser and less noticeable (my machine slowed down every
  time exmh did an inc with the perl version).

  Cheers to Jason for writing such a useful program. I hope this can
  be useful too.

  */

/* By this point, I (Jason Rennie) have made quite a few changes to the
   original port by Diego.  His code for reading in the database and
   lexing the messages has been only minimally changed, however.
   */
  

#include <sys/wait.h>     /* wait() */
#include <math.h>         /* log() */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <stdarg.h>       /* va_*() */
#include <assert.h>       /* assert() */
#include <unistd.h>       /* pipe() fork() close() dup2() execlp() */
#include <errno.h>	  /* required by Solaris 5.5.1/gcc-2.7.2 for errno variable */
#include <ctype.h>        /* tolower() */
#include <string.h>

#include "assoc_array_int.h"

#include "ifile.h"        /* standard ifile library */
#include "error.h"        /* ifile message and error reporting library */
  
FILE * INFO = NULL;

/* Main program */
int main(int argc, char **argv) {
  /* To the extent that it was possible, the original variable names
     were kept to avoid confusion. */
  clock_t DMZ_start, DMZ_end;
  int i,j;
  char *home_dir = NULL;
  char tmpstr[BUFSIZ*8], *tmpstrptr = NULL;
  char *acc_file = NULL;
  int tot_words;
  extendable_array tv;
  char *data_file = NULL;
  FILE *DATA = NULL;
  int tvsize;
  assoc_array_int n, docs, index, age, nk;
  int tot_docs;
  int tmpint;
  extendable_array word;
  char *curword = NULL;
  int cat_num, freq;
  char tmpstr2[BUFSIZ];
  float best_rating;
  int best_target;
  extendable_array lines;
  float target_rating;
  char nonalpha[256];
  char *wrd = NULL;
  extendable_array words;
  int word_count;
  float nval, docval;
  mh_profile_struct mh_profile;

  if (argc > 1)
    ifilter_print_usage();

  INFO = ifile_open_log(argc, argv);

  /* Get home directory */
  home_dir = getenv("HOME");
  if (home_dir == NULL)
    ifile_error("Fatal: HOME environment variable not defined!\n");
  ifile_verbosify(ifile_verbose, "home directory = %s\n", home_dir);

  /* Reads in info from ~/.mh_profile - currently not used for much */
  mh_profile = ifile_read_profile(home_dir);

  /* In the future, we should not update the accuracies file until it is
     guaranteed that a file is being presented at standard input */
  acc_file = ifile_sprintf("%s/%s", home_dir, IFILE_ACCURACY);
  ifile_update_accuracy(acc_file, 1, 0);

  
  /* Now read the database */
  tot_words = 0;
  
  EXT_ARRAY_INIT(tv,char*,EXT_ARRAY_CHUNK_SIZE);
  EXT_ARRAY_SET(tv,char*,0,NULL);
  tvsize = 0;

  data_file = ifile_sprintf("%s/.idata", home_dir);
  
  DMZ_start=clock();
  ifile_verbosify(ifile_progress, "Opening %s for reading...\n", data_file);

  DATA = fopen(data_file, "r");
  if (DATA != NULL)
    {
      /* Read the category names */
      readline(tmpstr, sizeof(tmpstr), DATA);
      tmpstrptr = strtok(tmpstr, " \t");
      while (tmpstrptr != NULL)
	{
	  tvsize++;
	  EXT_ARRAY_SET(tv, char *, tvsize, strdup(tmpstrptr));
	  tmpstrptr = strtok(NULL, " \t");
	}

      /* Read total number of words for each category */
      readline(tmpstr, sizeof(tmpstr), DATA);
      assoc_array_int_init(n);
      i = 1;
      tmpstrptr = strtok(tmpstr, " \t");
      while (tmpstrptr != NULL)
	{
	  assoc_array_int_set(n, EXT_ARRAY_GET(tv,char*,i), atoi(tmpstrptr));
	  i++;
	  tmpstrptr = strtok(NULL, " \t");
	}

      if (i <= tvsize)
	ifile_error("Bad data file format - line #2");
      
      /* Read number of documents for each category */
      readline(tmpstr, sizeof(tmpstr), DATA);
      assoc_array_int_init(docs);
      tot_docs = 0;
      i = 1;
      tmpstrptr = strtok(tmpstr, " \t");
      while (tmpstrptr != NULL)
	{
	  tmpint = atoi(tmpstrptr);
	  assoc_array_int_set(docs, EXT_ARRAY_GET(tv,char*,i), tmpint);
	  i++;
	  tot_docs += tmpint;
	  tmpstrptr = strtok(NULL, " \t");
	}
      if(i<=tvsize)
	ifile_error("Bad data file format - line #3");
      
      /* Read word frequencies */
      EXT_ARRAY_INIT(word,char*,8000);
      EXT_ARRAY_SET(word,char*,0,NULL);
      assoc_array_int_init(index);
      assoc_array_int_init(age);
      assoc_array_int_init(nk);
      
      readline(tmpstr, sizeof(tmpstr), DATA);
      while (!feof(DATA))
	{
	  tmpstrptr = strtok(tmpstr, " \t");
	  if (tmpstrptr == NULL)
	    break;
	  tot_words++;
	  EXT_ARRAY_SET(word, char *, tot_words, strdup(tmpstrptr));
	  assoc_array_int_set(index, tmpstrptr, tot_words);
	  curword = tmpstrptr;
	  
	  /* Read age of word */
	  tmpstrptr = strtok(NULL, " \t");
	  assoc_array_int_set(age, curword, atoi(tmpstrptr));
	  
	  /* Read frequency per category cat_num:freq */
	  tmpstrptr = strtok(NULL, ":");
	  while (tmpstrptr != NULL)
	    {	
	      cat_num = atoi(tmpstrptr);
	      tmpstrptr = strtok(NULL, " \t");
	      freq = atoi(tmpstrptr);
	      sprintf(tmpstr2, "%s,%s",
		      EXT_ARRAY_GET(word, char*, tot_words),
		      EXT_ARRAY_GET(tv, char*, cat_num));
	      assoc_array_int_set(nk, tmpstr2, freq);
	      
	      tmpstrptr = strtok(NULL, ":");
	    }
	  readline(tmpstr, sizeof(tmpstr), DATA);
	}
    }
  else
    {
      /* If no .idata is found, just pass to rcvstore */
      EXT_ARRAY_INIT(lines, char*, EXT_ARRAY_CHUNK_SIZE);
      i = 0;
      
      /* Read the message in an array for processing */
      readline(tmpstr, sizeof(tmpstr), stdin);
      while(!feof(stdin))
	{
	  EXT_ARRAY_SET(lines, char*, i, tmpstr);
	  i++;
	  readline(tmpstr, sizeof(tmpstr), stdin);
	}
      
      ifile_rcvstore(lines, "inbox");
      exit(0);
    }
  
  fclose(DATA);

  DMZ_end = clock();
  ifile_verbosify(ifile_progress, "Read %d categories, %d words.  Elapsed time: %.3f sec\n", tvsize, tot_words,
		  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);

  DMZ_start = clock();
  ifile_verbosify(ifile_progress, "Reading message...\n");

  /* Initialize the non-alphabetic-character token separator */
  j = 0;
  for(i = 1; i <= 255; i++)
    {
      if(!((i >= 'a' && i <= 'z') || (i >= 'A' && i <= 'Z')))
	nonalpha[j++] = i;
    }
  nonalpha[j] = '\0';
  
  EXT_ARRAY_INIT(lines,char*,EXT_ARRAY_CHUNK_SIZE);
  EXT_ARRAY_INIT(words,char*,500);
  i = 0;
  word_count = 0;
  
  /* Read the message in an array for processing */
  readline(tmpstr, sizeof(tmpstr), stdin);
  while (!feof(stdin))
    {
      EXT_ARRAY_SET(lines, char*, i, strdup(tmpstr));
      i++;
      /* Split line in words and store them in another array */
      tmpstrptr = strtok(tmpstr,nonalpha);
      while (tmpstrptr != NULL)
	{
	  wrd = tmpstrptr;
	  /* Skip tiny words */
	  if (strlen(wrd)>2)
	    {
	      /* Convert to lower case */
	      while ((*wrd) != '\0')
		{
		  (*wrd) = tolower(*wrd);
		  wrd++;
		}
	      EXT_ARRAY_SET(words, char*, word_count, strdup(tmpstrptr));
	      word_count++;
	    }
	  tmpstrptr = strtok(NULL, nonalpha);
	}
      readline(tmpstr, sizeof(tmpstr), stdin);
    }
  DMZ_end = clock();
  ifile_verbosify(ifile_progress, "Finishing reading message.  Elapsed time: %.3f sec\n",
		  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);


  /* Suggestion by Chris Brown for changes in analysis loop:

     # Rate each "target" folder
     while (($tvkey, $tvval) = each %TV)
       {
         $target_rating = 0;
         foreach $word (WORDLIST)
           {
             $target_rating += log(($NK{$word.",".$tvval}+1) /
             ($N{$tvval} + $wordsize));
           }
         print TMP "$tvval given rating of $target_rating\n";
         print "($tvval, $tvkey) given rating of $target_rating\n";
         $target_rating += log(($DOCS{$tvval} + 1) / ($docsize + $tvsize));
     
         # pick the best rating...
         if (($target_rating > $best_rating) || ($best_target == 0))
	   {
             $best_rating = $target_rating;
             $best_target = $tvval;
           }
       }
     */


  DMZ_start = clock();
  ifile_verbosify(ifile_progress, "Computing category ratings...\n");
  
  docval = (float)(tot_docs+tvsize);

  best_rating = 0.0;
  best_target = 0;
  
  /* Now do the analysis */
  for(i = 1; i <= tvsize; i++)
    {
      /* We don't want to filter into the draft folder */
      if (strcmp(mh_profile.draft_folder,
		 EXT_ARRAY_GET(tv, char *, i)) == 0)
	continue;

      target_rating = 0.0;
      tmpstrptr = EXT_ARRAY_GET(tv,char*,i);
      nval = (float)(assoc_array_int_get(n,tmpstrptr)+tot_words);
      for (j = 0; j < word_count; j++)
	{
	  wrd = EXT_ARRAY_GET(words,char*,j);
	  sprintf(tmpstr2, "%s,%s", wrd, tmpstrptr);
	  target_rating += log(((float)(assoc_array_int_get(nk,tmpstr2)+1))/
			       nval);
	}
    
      ifile_verbosify(ifile_debug, "%s given rating of %f\n", tmpstrptr,
		      target_rating);

      target_rating += log((float)(assoc_array_int_get(docs,tmpstrptr)+1)/
			   docval);
      if((target_rating > best_rating) || (best_target == 0))
	{
	  best_rating = target_rating;
	  best_target = i;
	}
    }

  DMZ_end=clock();
  ifile_verbosify(ifile_progress, "Chose %s (%d) as destination folder.  Elapsed time: %.3f sec\n",
		  EXT_ARRAY_GET(tv, char *, best_target), best_target,
		  ((double)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
  
  ifile_rcvstore(lines, EXT_ARRAY_GET(tv, char *, best_target));
  
  ifile_close_log();

  return 0;
}
